#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup


setup(name='threadpoolctl',
      version='3.0.0',
      description='threadpoolctl',
      author='Thomas Moreau',
      author_email='thomas.moreau.2010@gmail.com',
      url='https://github.com/joblib/threadpoolctl',
      py_modules=['threadpoolctl'],
      python_requires='>=3.6',
     )
